/*
 * cc.redberry.pipe: java library for implementation of concurrent pipelines
 * Copyright (c) 2010-2012
 * Bolotin Dmitriy <bolotin.dmitriy@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package cc.redberry.pipe;


import cc.redberry.pipe.util.SimpleProcessorWrapper;
import cc.redberry.pipe.util.SimpleProcessorWrapperSynchronized;
import cc.redberry.pipe.util.SimpleVoidProcessorExecutor;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

/**
 * Utility methods.
 */
public class CUtils {
    /**
     * Wraps processor and input port into single output port. Wrapper will introduce additional synchronization if
     * processor is not implementing {@link ThreadSafe} interface.
     *
     * @param inputPort output port to be wrapped
     * @param processor processor to wrap
     * @param <InputT>  type of input objects
     * @param <OutputT> type of output objects
     * @return wrapped output port
     */
    public static <InputT, OutputT> OutputPort<OutputT> wrap(OutputPort<? extends InputT> inputPort, Processor<? super InputT, ? extends OutputT> processor) {
        if (processor instanceof ThreadSafe)
            return new SimpleProcessorWrapper<>(inputPort, processor);
        else
            return new SimpleProcessorWrapperSynchronized<>(inputPort, processor);
    }

    /**
     * Returns void processor wrapper of an input port.
     *
     * @param inputPort input port to wrap
     * @param <T>       type of objects
     * @return wrapped input port
     */
    public static <T> VoidProcessor<T> asVoidProcessor(final InputPort<T> inputPort) {
        return new VoidProcessor<T>() {
            @Override
            public void process(T input) {
                try {
                    inputPort.put(input);
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                }
            }
        };
    }

    /**
     * Drains all content of output port into input port.<br/> This method will close a to port by putting null into
     * it.
     *
     * @param from port to take elements from
     * @param to   port to put elements to
     * @param <T>  type of elements
     * @throws InterruptedException
     */
    public static <T> void drain(OutputPort<T> from, InputPort<T> to) throws InterruptedException {
        T element;
        while ((element = from.take()) != null)
            to.put(element);
        to.put(null); //Closing
    }

    /**
     * Drains all content of output port into input port.
     *
     * @param from port to take elements from
     * @param to   port to put elements to
     * @param <T>  type of elements
     * @throws InterruptedException
     */
    public static <T> void drainWithoutClose(OutputPort<T> from, InputPort<T> to) throws InterruptedException {
        T element;
        while ((element = from.take()) != null)
            to.put(element);
    }

    /**
     * Converts VoidProcessor to Processor
     *
     * @param processor VoidProcessor to convert
     * @return Processor wrapper of VoidProcessor
     */
    public static <T> Processor<T, T> asProcessor(final VoidProcessor<T> processor) {
        return new Processor<T, T>() {
            @Override
            public T process(T input) {
                processor.process(input);
                return input;
            }
        };
    }

    /**
     * Process no more than {@code maxCount} elements from input by provided processor in current thread.
     *
     * @param input     input object stream
     * @param processor void processor
     * @param maxCount  maximum number of objects to process
     * @param <T>       type of processed objects
     * @return true if some objects left in input, false if not
     * @throws InterruptedException
     */
    public static <T> boolean processPart(OutputPort<T> input, VoidProcessor<T> processor, int maxCount) throws InterruptedException {
        if (maxCount <= 0)
            throw new IllegalArgumentException("maxCount should be greater than 0");
        T object = null;
        while ((--maxCount) >= 0 && (object = input.take()) != null)
            processor.process(object);
        return object != null;
    }

    /**
     * Process all elements from output port in current thread.
     *
     * @param input     input object stream
     * @param processor void processor
     * @param <T>       type of processed objects
     * @throws InterruptedException
     */
    public static <T> void processAll(OutputPort<T> input, VoidProcessor<T> processor) throws InterruptedException {
        T object;
        while ((object = input.take()) != null)
            processor.process(object);
    }

    /**
     * Process all elements from output port in current thread.
     *
     * @param input       input object stream
     * @param processor   void processor
     * @param threadCount number of parallel threads
     * @param <T>         type of processed objects
     * @throws InterruptedException
     */
    public static <T> void processAllInParallel(OutputPort<T> input, VoidProcessor<T> processor, int threadCount)
            throws InterruptedException {
        if (!(processor instanceof ThreadSafe) && threadCount > 1)
            throw new IllegalArgumentException("Using non-threadsafe VoidProcessor in several threads.");

        --threadCount;
        Thread[] threads = new Thread[threadCount];
        for (int i = 0; i < threadCount - 1; ++i) {
            threads[i] = new Thread(new SimpleVoidProcessorExecutor<>(input, processor));
            threads[i].start();
        }
        T object;
        while ((object = input.take()) != null)
            processor.process(object);
        for (int i = 0; i < threadCount; ++i)
            threads[i].join();
    }

    /**
     * Process all elements from output port in current thread.
     *
     * @param input           input object stream
     * @param processor       void processor
     * @param threadCount     number of parallel threads
     * @param executorService executor service to create threads
     * @param <T>             type of processed objects
     * @throws InterruptedException
     */
    public static <T> void processAllInParallel(OutputPort<T> input, VoidProcessor<T> processor, int threadCount,
                                                ExecutorService executorService) throws InterruptedException {
        if (!(processor instanceof ThreadSafe) && threadCount > 1)
            throw new IllegalArgumentException("Using non-threadsafe VoidProcessor in several threads.");

        --threadCount;
        Future<?>[] futures = new Future[threadCount];
        for (int i = 0; i < threadCount; ++i)
            futures[i] = executorService.submit(new SimpleVoidProcessorExecutor<>(input, processor));
        T object;
        while ((object = input.take()) != null)
            processor.process(object);
        try {
            for (int i = 0; i < threadCount; ++i)
                futures[i].get();
        } catch (ExecutionException ex) {
            throw (RuntimeException) ex.getCause();
        }
    }

    /**
     * Process all elements from output port in current thread.
     *
     * @param input             input object stream
     * @param processorsFactory void processor factory
     * @param threadCount       number of parallel threads
     * @param <T>               type of processed objects
     * @throws InterruptedException
     */
    public static <T> void processAllInParallel(OutputPort<T> input, VoidProcessorFactory<T> processorsFactory, int threadCount)
            throws InterruptedException {
        --threadCount;
        Thread[] threads = new Thread[threadCount];
        for (int i = 0; i < threadCount; ++i) {
            threads[i] = new Thread(new SimpleVoidProcessorExecutor<T>(input, processorsFactory.create()));
            threads[i].start();
        }

        VoidProcessor<T> processor = processorsFactory.create();
        T object;
        while ((object = input.take()) != null)
            processor.process(object);

        for (int i = 0; i < threadCount; ++i)
            threads[i].join();
    }

    /**
     * Processes elements in N different threads. For each thread a separate processor is created by a provided
     * factory.<br/> In this implementation threads are created by an ExecutorsService.
     *
     * @param input             elements to process
     * @param processorsFactory factory to createInputStream processors
     * @param threadCount       count of parallel processing threads
     * @param executorService   service to execute this task (thread pool)
     * @param <T>               type of processed objects
     */
    public static <T> void processAllInParallel(OutputPort<T> input, VoidProcessorFactory<T> processorsFactory, int threadCount,
                                                ExecutorService executorService) throws InterruptedException {
        --threadCount;
        Future<?>[] futures = new Future[threadCount];
        for (int i = 0; i < threadCount; ++i)
            futures[i] = executorService.submit(new SimpleVoidProcessorExecutor<>(input, processorsFactory.create()));

        VoidProcessor<T> processor = processorsFactory.create();
        T object;
        while ((object = input.take()) != null)
            processor.process(object);
        try {
            for (int i = 0; i < threadCount; ++i)
                futures[i].get();
        } catch (ExecutionException ex) {
            throw (RuntimeException) ex.getCause();
        }
    }
}
